#!/usr/bin/env python3
"""
This module contains functions to segment the two phases of a bicontinuous
nanostructured material.  Phases can be solid/pore (nanoporous metals) or
solid/solid (bicontinuous composites).
"""
import warnings

import time
import numpy as np
import matplotlib.pyplot as plt
from scipy import ndimage
from skimage.filters import threshold_otsu, rank
from skimage.morphology import disk, remove_small_objects
from skimage.restoration import denoise_bilateral
from skimage.color import rgb2gray

from aquami import inout
from aquami import display
from aquami import gui

__author__ = "Joshua Stuckner"

def smoothEdges(mask, smooth_radius=1):
    """
    Smoothes the edges of a binary mask.

    Parameters
    ----------
    mask : 2D array of bool
        Mask showing the area of one phase.
    smooth_radius : float64
        The radius of the smoothing operation.  See note below.
        
    Returns
    -------
    smooth_mask : 2D array of bool
        Mask that has been smoothed.

    Notes
    -----
    smooth_radius sets the structure element (selem) for smoothing the edges of
    the masks. If the smooth_rad rounds up the selem is a disk with radius
    rounded up.  If smooth_radius rounds down, selem is a box.

    Radius = 0 - 1.499
    [[1,1,1],
     [1,1,1],
     [1,1,1]]

    Radius = 1.5 - 1.99
    [[0,0,1,0,0],
     [0,1,1,1,0],
     [1,1,1,1,1],   
     [0,1,1,1,0],
     [0,0,1,0,0]]
     
    Radius = 2 - 2.499
    [[1,1,1,1,1],
     [1,1,1,1,1],
     [1,1,1,1,1],   
     [1,1,1,1,1],
     [1,1,1,1,1]]    
    """

    smooth_radius = max(smooth_radius, 1)

    if round(smooth_radius, 0) > int(smooth_radius): # If round up.
        size = int(smooth_radius + 1)
        selem = disk(round(smooth_radius,0))
    else:
        size = 1 + 2*int(smooth_radius)
        selem = np.ones((size,size))

    # Smooth edges.
    # It is necessary to perform this step because skelatonizing algorithms are
    # extremely sensitive to jagged edges and may otherwise give spurious
    # results.
    smooth_mask = ndimage.binary_opening(mask, structure=selem)
    smooth_mask = ndimage.binary_closing(smooth_mask, structure=selem)

    return smooth_mask

def removeSmallObjects(mask, small_objects=0, small_holes=0):
    """
    Removes small objects (white areas of mask) and small holes (black areas).

    Parameters
    ----------
    mask : 2D array of bool
        Mask showing the area of one phase.
    small_objects : int
        Max area of connected white pixels that will be removed.
    small_holes : int
        Max area of connected black pixels that will be removed.        
        
    Returns
    -------
    out_mask : 2D array of bool
        Mask with small holes and objects removed.
    """

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        out_mask = remove_small_objects(mask, small_objects)
        out_mask = ~remove_small_objects(~out_mask, small_holes)

    return out_mask
        
def roughSegment(image):
    """
    Returns binary masks of the segmented bright and dark phases.  Uses Otsu's
    Method and some morphological operations to reduce noise.

    Parameters
    ----------
    image : 2D array of uint8
        Image data to be segmented

    Returns
    -------
    bright : 2D array of bool
        Mask of bright phase.
    dark : 2D array of bool
        Mask of dark phase or pores.
    """

    # Don't operate on the passed image.
    img = np.copy(image)
    
    # Get shape of image and convert to grayscale.
    try:
        rows, cols =  img.shape
    except ValueError: # Convert to grayscale
        img = rgb2gray(img)
        rows, cols =  img.shape

    # Ensure proper datatype.
    img = inout.uint8(img)

    # Set radius of smoothing filter based on the image resolution.
    blur_mult = 1.4 # Increasing this parameter increases the blur radius.
    blur_radius = max(1,(cols*blur_mult/1024))

    # Smooth the image to reduce noise.
    img = ndimage.gaussian_filter(img, blur_radius)

    # Ensure proper datatype after bilateral smoothing.
    img = inout.uint8(img)
    
    # Use Otsu's Method to determine the global threshold value.
    threshold_global_otsu = threshold_otsu(img)
    # Generate masks of both phases based on Otsu's threshold value.
    bright = img >= threshold_global_otsu
    dark = img < threshold_global_otsu

    # Morphological operations to remove noise.
    small = 200 # Area in pixels^2 of objects to remove as noise
    bright = smoothEdges(bright, 1)
    dark = smoothEdges(dark, 1)
    bright = removeSmallObjects(bright, small, small)
    dark = removeSmallObjects(dark, small, small)

    return bright, dark

def segment(image, estSize):
    """
    Returns binary masks of the segmented bright and dark phases.  Uses 
    two-dimentional Otsu's Method and some morphological operations to
    reduce noise.  2D Otsu's Method takes significantly longer to perform than
    the regular method and is more robust against illumination and contrast
    changes throughout the image by setting a seperate threshold value for each
    pixel.

    Parameters
    ----------
    image : 2D array of uint8
        Image data to be segmented
    estSize : float
        Rough estimate of ligament diameter.  Used to set parameters such as
        the blur radius for smoothing.

    Returns
    -------
    bright : 2D array of bool
        Mask of bright phase.
    dark : 2D array of bool
        Mask of dark phase or pores.
    """

    # Don't operate on the passed image.
    img = np.copy(image)

    # Adjustable parameters.
    # These have been highly tested and changing these may throw off results.
    otsuMultiplier = 5 # Otsu neighborhood = otsuMultiplier * estSize
    blur_mult = 1.4 # Increasing this parameter increases the blur radius.
    small_mult = 0.025 # Remove objects with area < small_mult*3.14*estSize^2.
    smooth_div = 36 # Sets the structure element radius for edge smoothing.
    small_blob_mult = 0.25 # Used to remove small blobs after edge smoothing.
    small_hole_mult = 0.05 # Used to remove small holes after edge smoothing.
    

    # Get shape of image and convert to grayscale.
    try:
        rows, cols =  img.shape
    except ValueError: # Convert to grayscale
        img = rgb2gray(img)
        rows, cols =  img.shape

     
    # Smooth the image.
    blur_radius = max(1,int(round(cols*blur_mult/1000,0)))
    img = denoise_bilateral(img,
                            sigma_color=0.05,  #increase this to blur more
                            sigma_spatial=blur_radius,
                            multichannel=False)
 
    # Ensure proper datatype after bilateral smoothing.
    img = inout.uint8(img)

    # Perform Otsu thresholding
    threshold_global_otsu = threshold_otsu(img)
    global_otsu = img >= threshold_global_otsu    

    # Perform 2D Otsu thresholding.
    radius = estSize * otsuMultiplier
    selem = disk(radius)
    local_otsu = rank.otsu(img, selem)
    bright = img >= local_otsu
    dark = img < local_otsu

    
    # Remove tiny objects in both phases.
    small = small_mult * 1/2 * 3.14 * estSize**2
    bright = removeSmallObjects(bright, small, small)
    dark = removeSmallObjects(dark, small, small)

    # Smooth edges.  
    smooth_rad = max(estSize/smooth_div, 1)
    bright = smoothEdges(bright, smooth_rad)
    dark = smoothEdges(dark, smooth_rad)

    # Remove small blobs and really small holes.
    small_blobs = small_blob_mult * 1/2 * 3.14 * estSize**2
    small_holes = small_hole_mult * 1/2 * 3.14 * estSize**2
    bright = removeSmallObjects(bright, small_blobs, small_holes)
    dark = removeSmallObjects(dark, small_blobs, small_holes)
    
    return bright, dark

def manualSegment(image):
    """
    Returns binary masks of the segmented bright and dark phases based on
    manually selected threshold value.

    Parameters
    ----------
    image : 2D array of uint8
        Image data to be segmented

    Returns
    -------
    bright : 2D array of bool
        Mask of bright phase.
    dark : 2D array of bool
        Mask of dark phase or pores.
    """

    # Adjustable parameters.
    # These have been highly tested and changing these may throw off results.
    blur_mult = 1.4 # Increasing this parameter increases the blur radius.
    small_div = 20 # Increasing this reduces hole close size in low res images.

    
    img = inout.uint8(image)

    # Get shape of image and convert to grayscale.
    try:
        rows, cols =  img.shape
    except ValueError: # Convert to grayscale
        img = rgb2gray(img)
        rows, cols =  img.shape

    small = int((min(rows,cols)/small_div)**2)
    if small > 200:
        small = 200

    # Smooth the image to reduce noise.
    blur_radius = max(1,(cols*blur_mult/1024))
    img = ndimage.gaussian_filter(img, blur_radius)

    # Get the default threshold value.
    threshinit = threshold_otsu(img)

    # Launch the GUI for manual threshholding.
    p = gui.manualThreshhold(img, threshinit)
    p.show()
    bright = p.getMask()
    dark = ~bright

    # Morphological operations to remove noise.
    bright = smoothEdges(bright, 1)
    dark = smoothEdges(dark, 1)
    bright = removeSmallObjects(bright, small, small)
    dark = removeSmallObjects(dark, small, small)

    return bright, dark



#!/usr/bin/env python3
"""
This module contains functions for displaying final and intermediate results
when computing segmentation and measurements of bicontinuous nanostructured
materials or dealloyed materials.
"""

import warnings

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import animation
import scipy.ndimage.morphology as morphology
import skimage.measure as measure
from skimage import img_as_ubyte
from scipy.stats import lognorm, norm
from skimage.color import gray2rgb, rgb2gray

__author__ = "Joshua Stuckner"
__all__ = ['showFull', 'showOtsu', 'showSkel', 'showHist', 'overlayMask']

def showFull(img, title=None, cmap=None, interpolation='none'):
    """
    Displays a full screen figure of the image.

    Parameters
    ----------
    img : ndarray
        Image to display.
    title : str, optional
        Text to be displayed above the image.
    cmap : Colormap, optional
        Colormap that is compatible with matplotlib.pyplot
    interpolation : string, optional
        How display pixels that lie between the image pixels will be handled.
        Acceptable values are ‘none’, ‘nearest’, ‘bilinear’, ‘bicubic’,
        ‘spline16’, ‘spline36’, ‘hanning’, ‘hamming’, ‘hermite’, ‘kaiser’,
        ‘quadric’, ‘catrom’, ‘gaussian’, ‘bessel’, ‘mitchell’, ‘sinc’, ‘lanczos’
    """

    # Show grayscale if cmap not set and image is not color.
    if cmap is None and img.ndim == 2:
        cmap = plt.cm.gray
        
    plt.imshow(img, cmap = cmap, interpolation=interpolation)
    plt.axis('off')
    figManager = plt.get_current_fig_manager()
    figManager.window.showMaximized()   
    if title is None:
        plt.gca().set_position([0, 0, 1, 1])
    else:
        plt.gca().set_position([0, 0, 1, 0.95])
        plt.title(title)
    plt.show()


def showOtsu(img, local_otsu, global_otsu, radius, threshold_global_otsu):
    """
    Displays an image and the results of segmentation via regular Otsu's
    method and 2D Otsu's Method along with the threshold value at each pixel
    for direct comparison.  Taken from scikit-image.org

    Parameters
    ----------
    img : image array
        Original image before segmentation.
    local_otsu : 2D array
        Array containing threshold value at each pixel determined by 2D Otsu's
        Method.
    global_otsu : 2D binary array
        Segmentation of the original image by Otsu's Method.
    radius : float
       Radius of neighborhood used for 2D Otsu's Method.
    threshold_global_otsu: int
       Global threshold value determined by Otsu's Method.
    """

    fig, ax = plt.subplots(2, 2, figsize=(8, 5), sharex=True,
                           sharey=True, subplot_kw={'adjustable':'box-forced'})
    ax1, ax2, ax3, ax4 = ax.ravel()

    fig.colorbar(ax1.imshow(img, cmap=plt.cm.gray),
                 ax=ax1, orientation='horizontal')
    ax1.set_title('Original')
    ax1.axis('off')

    fig.colorbar(ax2.imshow(local_otsu, cmap=plt.cm.gray),
                 ax=ax2, orientation='horizontal')
    ax2.set_title('Local Otsu (radius=%d)' % radius)
    ax2.axis('off')

    ax3.imshow(img >= local_otsu, cmap=plt.cm.gray)
    ax3.set_title('Original >= Local Otsu' % threshold_global_otsu)
    ax3.axis('off')

    ax4.imshow(global_otsu, cmap=plt.cm.gray)
    ax4.set_title('Global Otsu (threshold = %d)' % threshold_global_otsu)
    ax4.axis('off')

    plt.axis('off')
    plt.show()
    plt.close()


def showSkel(skeleton, mask, dialate=False, title=None, returnSkel=False,
             cmap=plt.cm.spectral):
    """
    Displays skelatal data on top of an outline of a binary mask. For example,
    displays a medial axis transform over an outline of segmented ligaments.

    Parameters
    ----------
    skeleton : 2D array
        Data to be displayed.
    mask : binary 2D array
        Mask of segmentation data, the outline of which is displayed along with
        the skel data.
    dialate : boolean, optional
        If dialate is true, the skelatal data will be made thicker in the
        display.
    title : str, optional
        Text to be displayed above the image.
    """

    skel = np.copy(skeleton)

    # Find the outlines of the mask and make an outline mask called outlines.
    contours = measure.find_contours(mask, 0.5)
    outlines = np.zeros((mask.shape), dtype='uint8')
    for n, contour in enumerate(contours):
        for i in range(len(contour)):
            outlines[int(contour[i,0]), int(contour[i,1])] = 255

    # Make the skel data thicker if dialate is true.
    if dialate:
        skel = morphology.grey_dilation(skel, size=(3,3))


    # Scale the skel data to uint8 and add the outline mask
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        skel = skel.astype(np.float32) # convert to float
        skel -= skel.min() # ensure the minimal value is 0.0
        if skel.max() != 0:
            skel /= skel.max() # maximum value in image is now 1.0
    skel = np.uint8(cmap(skel)*255) # apply colormap to skel data.
    for i in range(3):
        skel[:,:,i] += outlines

    if returnSkel:
        return skel
    
    # Display the results.
    plt.imshow(skel, cmap = cmap, interpolation='none')
    plt.axis('off')
    figManager = plt.get_current_fig_manager()
    figManager.window.showMaximized()
    if title is None:
        plt.gca().set_position([0, 0, 1, 1])
    else:
        plt.gca().set_position([0, 0, 1, 0.95])
        plt.title(title)
    plt.show()


def showHist(data, title=None, xlabel=None,
             numBins=None, gauss=False, log=False):
    """
    Displays a histogram of data with no y-axis and the option to fit gaussian
    and lognormal distribution curves to the data.

    Parameters
    ----------
    data : ndarray
        Data or measurements from which to produce a histogram.
    title : str, optional
        Title of histogram which is displayed at the top of the figure.
    xlabel : str, optional
        Title of the x-axis.  Usually what is measured along the x-axis along
        with the units.
    numBins : int, optional
        Number of possible bars in the histogram.  If not given, the function
        will attempt to automatically pick a good value.
    gauss: boolean, optional
        If true, a fitted guassian distribution curve will be plotted on the
        histogram.
    log: boolean, optional
        If true, a fitted lognormal distribution curve will be plotted on the
        histogram.
    """

    # Determine optimum number of bins.
    if numBins is None:
        u = len(np.unique(data))
        numBins = int(2*u**(1/2))
        if numBins < 4:
            numBins = len(np.unique(data))
            if numBins < 1:
                numBins = 1

    # Create the histogram.
    n, bins, patches = plt.hist(data, bins=numBins, normed=1, edgecolor='black')
    if log:
        try:
            logfit = lognorm.fit(data.flatten(), floc=0)
            pdf_plot = lognorm.pdf(bins, 
                                   logfit[0], loc=logfit[1], scale=logfit[2])
            plt.plot(bins, pdf_plot, 'r--', linewidth=3, label='lognorm')
        except ValueError:
            pass
    if gauss:
        gfit = norm.fit(data.flatten())
        gauss_plot = norm.pdf(bins, gfit[0], gfit[1])
        plt.plot(bins, gauss_plot, 'g--', linewidth=3, label='gaussian')

    # Display the histogram.
    if title is not None:
        plt.title(title)
    if xlabel is not None:
        plt.xlabel(xlabel)
    ax = plt.gca()
    ax.yaxis.set_visible(False)
    if gauss or log:
        plt.legend()
    plt.show()
  
def overlayMask(image, mask, color='o', return_overlay=False, animate=False,
                title=None):
    '''
    Displays the binary mask over the original image in order to verify results.
    
    Parameters
    ----------
    image : image array
        Image data prior to segmentation.
    mask : binary array
        Binary segmentation of the image data.  Must be the same size as image.
    color : str, optional
        The color of the overlaid mask.
    return_overlay : bool, optional
        If true, the image with the overlaid mask is returned and the overlay
        is not displayed here.
    animate : bool, optional
        If true, an animated figure will be displayed that alternates between
        showing the raw image and the image with the overlay.

    Returns
    -------
    overlay : RGB image array, optional
        Color image with mask overlayyed on original image (only returned
        if 'return_overlay' is True).
    '''

    if title is None:
        title = 'Segmentation mask overlayed on image'

    img = np.copy(image)

    # Convert the image into 3 channels for a colored mask overlay
    overlay = gray2rgb(img)

    # Set color (default to blue if a proper color string is not given).
    r = 0
    g = 0
    b = 255
    if color == 'red' or color == 'r':
        r = 255
        g = 0
        b = 0
    if color == 'green' or color == 'g':
        r = 0
        g = 255
        b = 0
    if color == 'blue' or color == 'b':
        r = 0
        g = 0
        b = 255
    if color == 'white' or color == 'w':
        r = 255
        g = 255
        b = 255
    if color == 'yellow' or color == 'y':
        r = 255
        g = 255
        b = 0
    if color == 'orange' or color == 'o':
        r = 255
        g = 128
        b = 0
        
    # Apply mask.
    if r != 0:
        overlay[mask == 1, 0] = r
    if g != 0:
        overlay[mask == 1, 1] = g
    if b != 0:
        overlay[mask == 1, 2] = b

    # Return or show overlay.
    if return_overlay:
        return overlay
    else:
        if animate:
            fig = plt.figure()
            ims = []
            ims.append([plt.imshow(image, cmap=plt.cm.gray, animated=True)])
            ims.append([plt.imshow(overlay, animated=True)])
            ani = animation.ArtistAnimation(fig, ims, 1000, True, 1000)
            plt.axis('off')
            figManager = plt.get_current_fig_manager()
            figManager.window.showMaximized()
            plt.gca().set_position([0, 0, 1, 0.95])
            plt.title(title)
            fig.canvas.set_window_title('Animated Mask Overlay')
            plt.show()            
        else:           
            showFull(overlay, title=title,interpolation='nearest')





#!/usr/bin/env python3
"""
This module contains GUI elements for the BNMImage package.
"""

import os
import sys

import tkinter as tk
from tkinter.scrolledtext import ScrolledText    
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from matplotlib.widgets import Slider, Button
from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_pdf import PdfPages
from skimage.color import rgb2gray, gray2rgb
from skimage.restoration import denoise_bilateral
from scipy.misc import imread, imsave
import time
import pandas as pd

import copy

from aquami import tkSimpleDialog
from aquami import display
from aquami import measure
from aquami import segment
from aquami import inout
from aquami import autoSelect

__author__ = "Joshua Stuckner"
__version__ = '1.0'

class selectRect(object):
    """
    Draws a rectange on a figure and keeps track of the rectangle's size and
    location.  Used to select the image data and scale bar.

    Attributes
    ----------
    x0 : float64
        X coordinate (row) of start of rectangle.
    y0 : float 64
        Y coordinate (column) of start of rectangle.
    x1 : float64
        X coordinate (row) of end of rectangle.
    y1 : float 64
        Y coordinate (column) of end of rectangle.
    """
    
    def __init__(self):
        self.ax = plt.gca()
        self.rect = Rectangle((0,0), 1, 1,
                              facecolor='none',
                              edgecolor='#6CFF33',
                              linewidth=3)
        self.x0 = None
        self.y0 = None
        self.x1 = None
        self.y1 = None
        self.ax.add_patch(self.rect)
        self.ax.figure.canvas.mpl_connect('button_press_event', self.on_press)
        self.ax.figure.canvas.mpl_connect('button_release_event',
                                          self.on_release)

    def on_press(self, event):
        self.x0 = event.xdata
        self.y0 = event.ydata

    def on_release(self, event):
        self.x1 = event.xdata
        self.y1 = event.ydata
        if (self.x1 is not None and self.x0 is not None and
                            self.y1 is not None and self.y0 is not None):
            self.rect.set_width(self.x1 - self.x0)
            self.rect.set_height(self.y1 - self.y0)
            self.rect.set_xy((self.x0, self.y0))
            self.ax.figure.canvas.draw()

class inputScale(tkSimpleDialog.Dialog):        
    def body(self, master):

        self.varUnit = tk.StringVar()
        self.varUnit.set("nm")

        self.directions = tk.Label(master,
                text="Please manually input the scale bar number.")

        self.textInput = tk.Text(master, width=8, height=1)
        self.optUnit = tk.OptionMenu(master, self.varUnit, "nm", "μm", "mm")
        
        f = Figure()
        a = f.add_subplot(111)
        a.imshow(self.scale, cmap=plt.cm.gray, interpolation=None)
        canvas = FigureCanvasTkAgg(f, master=master)
        canvas.show()

        
        canvas.get_tk_widget().grid(row=1, column=0, columnspan=2)
        canvas._tkcanvas.grid(row=1, column=0, columnspan=2)

        self.directions.grid(row=0, column=0, columnspan=2)

        self.textInput.grid(row=2, column=0)
        self.optUnit.grid(row=2, column=1)
 

        return self.textInput #initial focus

    def apply(self):
        mult = 1
        if self.varUnit.get() == "μm":
            mult = 1000
        elif self.varUnit.get() == "mm":
            mult = 1000000
        self.result = int(int(self.textInput.get("1.0",tk.END))*mult)


class manualThreshhold(object):
    def __init__(self, img, threshinit):
        self.img = img
        self.original = img.copy()
        self.img = denoise_bilateral(self.img,
                            sigma_color=0.05,  #increase this to blur more
                            sigma_spatial=1,
                            multichannel=False)
        self.fig, self.ax = plt.subplots()
        plt.subplots_adjust(left=0.01, bottom=0.20)
        self.manThresh = 0
        self.pltimage = self.ax.imshow(self.img, interpolation='nearest')
        self.ax_thresh = plt.axes([0.2, 0.1, 0.65, 0.03])
        self.s_thresh = Slider(self.ax_thresh, 'Threshold', 0, 255,
                               valinit=threshinit)
        self.s_thresh.on_changed(self.threshUpdate)
        self.threshUpdate(val=1)
        mng = plt.get_current_fig_manager()
        mng.window.showMaximized()
        plt.axis('off')       

    def threshUpdate(self, val):      
        val = self.s_thresh.val
        self.manThreshVal = self.s_thresh.val
        img = inout.uint8(self.img)
        self.mask = img >= int(val)
        overlay = display.overlayMask(self.original, self.mask, color='o',
                                      return_overlay=True)
        self.pltimage.set_data(overlay)        
        self.fig.canvas.draw_idle()

    def show(self):
        plt.show()

    def getMask(self):
        return self.mask

        
def selectFile(title="Select image", initialdir=None, multiple=False):
    """
    Launches a dialog to select a file.

    Parameters
    ----------
    initialdir : str, optional
        The start path of the select a file dialog.

    multiple : bool, optional
        If true, allows the selecting of multiple files.
        
    Returns
    -------
    file : str
        The directory and name of the selected file.
    """
    file = tk.filedialog.askopenfilename(
            initialdir=initialdir,
            multiple=multiple,
            title=title
            )
    return file


def selectFolder():
    """
    Launches a dialog to select a folder.

    Returns
    -------
    directory : str
        The directory of the selected folder.
    """   
    directory = tk.filedialog.askdirectory(
            title='Select file'
            )
    return directory


def saveFile(ftype=None, title='Save file'): 
    file = tk.filedialog.asksaveasfile(
            defaultextension=ftype,
            title=title
            )
    
    path = file.name
    file.close()
    return path


def manualSelectImage(img):
    '''
    Used to select the image data and the scale bar in a raw microscopy image
    that may contain other metadata besides the actual image.
    
    Parameters
    ----------
    img : ndarray
        Raw microscopy image.

    Returns
    -------
    im_data : ndarray
        Selected area containing just the image data

    scale : ndarray
        Selected area containing an image of the scale bar.
    '''

    # Get shape of image and convert to gray scale.
    try:
        rows, cols =  img.shape
    except ValueError: # Convert to gray scale
        img = rgb2gray(img)
        rows, cols =  img.shape
    
    # Select the image.
    plt.imshow(img, cmap=plt.get_cmap('gray'))
    plt.title('Select the image data and close this figure.')
    figManager = plt.get_current_fig_manager()
    figManager.window.showMaximized()
    a = selectRect()
    plt.axis('off')
    plt.show()
    # Make sure the selection isn't out of bounds or reversed.
    y0 = min(a.y0, a.y1)
    y1 = max(a.y0, a.y1)
    x0 = min(a.x0, a.x1)
    x1 = max(a.x0, a.x1)
    y0 = 0 if y0 < 0 else y0
    y1 = rows if y1 > rows else y1
    x0 = 0 if x0 < 0 else x0
    x1 = cols if x1 > cols else x1
    # Crop the raw image to the image data.
    im_data = img[int(y0):int(y1), int(x0):int(x1)]
    
    # Select the scale bar.
    plt.imshow(img, cmap=plt.get_cmap('gray'))
    plt.title('Draw a box around the scale bar and close this figure.' \
              ' No need to be precise as long as the whole bar is included.')
    figManager = plt.get_current_fig_manager()
    figManager.window.showMaximized()
    a = selectRect()
    plt.axis('off')
    plt.show()
    # Make sure the selection isn't out of bounds or reversed.
    y0 = min(a.y0, a.y1)
    y1 = max(a.y0, a.y1)
    x0 = min(a.x0, a.x1)
    x1 = max(a.x0, a.x1)
    y0 = 0 if y0 < 0 else y0
    y1 = cols if y1 > cols else y1
    x0 = 0 if x0 < 0 else x0
    x1 = rows if x1 > rows else x1
    # Crop the raw image to the scale bar.   
    scale = img[int(a.y0):int(a.y1), int(a.x0):int(a.x1)]

    return im_data, scale


class CreateToolTip(object):
    '''
    create a tooltip for a given widget
    '''
    def __init__(self, widget, text='widget info'):
        self.widget = widget
        self.text = text
        self.widget.bind("<Enter>", self.enter)
        self.widget.bind("<Leave>", self.close)
        self.timer = []
        self.tw = None
        
    def enter(self, event=None):
        self.timer = self.widget.after(700, self.display)
        
    def display(self):
        x = y = 0
        x, y, cx, cy = self.widget.bbox("insert")
        x += self.widget.winfo_rootx() + 25
        y += self.widget.winfo_rooty() + 20
        # creates a toplevel window
        self.tw = tk.Toplevel(self.widget)
        # Leaves only the label and removes the app window
        self.tw.wm_overrideredirect(True)
        self.tw.wm_geometry("+%d+%d" % (x, y))
        label = tk.Label(self.tw, text=self.text, justify='left',
                       background='#ffffe6', relief='solid', borderwidth=1,
                       font=("times", "10", "normal"))
        label.pack(ipadx=5, ipady=3)
        
    def close(self, event=None):
        if self.tw is not None:
            self.tw.destroy()
        self.widget.after_cancel(self.timer)

class Aquami_Gui(tk.Frame):
    """
    This class in a GUI for BNMImage.
    """
    
    def __init__(self, master=None):
        """
        Called when class is created.
        """
        
        tk.Frame.__init__(self, master)
        self.master.title('AQUAMI')
        module_path, this_filename = os.path.split(__file__)
        self.master.iconbitmap(''.join((module_path, '/icon.ico')))

        self.initGUI()

    def initGUI(self):
        """
        Create and layout all the widgets.
        """

        self.pack(fill=tk.BOTH, expand=True)

        # Figure out sizing.
        width = 200
        height = 200
        pad = 5
        fontWidth = 8
        bigWidth = int((width*3 + pad*6) / fontWidth)
        
        # Create option frames.
        self.frameOptions = tk.LabelFrame(self, text="Options:",
                                          width=width, height=height)
        self.frameSegment = tk.LabelFrame(self, text="Segmentation Method:",
                                          width=width, height=height)
        self.frameMeasure = tk.LabelFrame(self, text="Measurements:",
                                          width=width, height=height)

        # Create text boxes and labels.
        self.labelStatus = tk.LabelFrame(self, text="Status:", bd=0)
        self.labelResults = tk.LabelFrame(self, text="Results:", bd=0)
        self.textStatus = ScrolledText(self.labelStatus, height=5,
                                       width=bigWidth)
        self.textResults = ScrolledText(self.labelResults, height=10,
                                        width=bigWidth)

        # Create buttons.
        self.buttonCalculate = tk.Button(self, text='Calculate',
                                         width=20, height=1, font=12, bd=3,
                                         command=lambda:self.prepare())
        self.buttonSaveAll = tk.Button(self, text='Save Session Summary',
                                       command=self.saveAll)
        self.buttonSelectOutFold = tk.Button(self, text='Set Output Folder',
                                             command=self.setOutputFolder)
        self.buttonAbout = tk.Button(self, text='About', command=self.about)

        # Arrange toplevel widgets.
        self.frameOptions.grid(row=0, column=2, padx=pad, pady=pad,
                               sticky='NESW')
        self.frameSegment.grid(row=0, column=1, padx=pad, pady=pad,
                               sticky='NESW')
        self.frameMeasure.grid(row=0, column=0, padx=pad, pady=pad,
                               sticky='NESW')

        self.buttonCalculate.grid(row=1, column=1, 
                                  padx=pad, pady=pad*3)
        self.buttonSelectOutFold.grid(row=1, column=0, 
                                  padx=pad, pady=pad*3)
        self.buttonAbout.grid(row=6, column=2, sticky='e', padx=20, pady=10)

        self.labelStatus.grid(row=2, column=0, columnspan=3, sticky='w',
                              padx=pad, pady=pad)
        self.textStatus.grid(row=3, column=0, columnspan=3)
        self.labelResults.grid(row=4, column=0, columnspan=3, sticky='w',
                               padx=pad, pady=pad)
        self.textResults.grid(row=5, column=0, columnspan=3)
        self.buttonSaveAll.grid(row=6, column=1, padx=pad, pady=pad)

        # Variables
        self.outFold = None
        columns = [["","","","",
                    "Bright phase diameter",
                    "","",
                    "Dark phase diameter",
                    "","",
                    "Bright length",
                    "","",
                    "Dark length",
                    "","",
                    "Bright area",
                    "","",
                    "Dark area",
                    "","",
                    "Bright connected length",
                    "","",
                    "Dark connected length",
                    "",""],                    
                   ["image",
                    "pixel size",
                    "area frac",
                    "est diam",
                    "Average",
                    "SD",
                    "Measured",
                    "Average",
                    "SD",
                    "Measured",
                    "Average",
                    "SD",
                    "Measured",
                    "Average",
                    "SD",
                    "Measured",
                    "Average",
                    "SD",
                    "Measured",
                    "Average",
                    "SD",
                    "Measured",
                    "Average",
                    "SD",
                    "Measured",
                    "Average",
                    "SD",
                    "Measured"]]
                   
        self.saveAll = np.array(columns)

        # Measurement options.
        # Variables.
        self.varDiameter = tk.BooleanVar()
        self.varLength = tk.BooleanVar()
        self.varArea = tk.BooleanVar()
        self.varSumConnectedLength = tk.BooleanVar()
        self.varAreaFraction = tk.BooleanVar()
        # Create widgets.
        self.checkDiameter = tk.Checkbutton(self.frameMeasure,
                text="Diameter", variable=self.varDiameter)
        self.checkLength = tk.Checkbutton(self.frameMeasure,
                text="Length", variable=self.varLength)
        self.checkArea = tk.Checkbutton(self.frameMeasure,
                text="Area", variable=self.varArea)
        self.checkSumConnectedLength = tk.Checkbutton(self.frameMeasure,
                text="Connected length", variable=self.varSumConnectedLength)
        self.checkAreaFraction = tk.Checkbutton(self.frameMeasure,
                text="Area fraction", variable=self.varAreaFraction)
        # Pack widgets.
        self.checkDiameter.grid(row=0, column=0, sticky='w')
        self.checkLength.grid(row=1, column=0, sticky='w')
        self.checkArea.grid(row=2, column=0, sticky='w')
        self.checkSumConnectedLength.grid(row=3, column=0, sticky='w')
        self.checkAreaFraction.grid(row=4, column=0, sticky='w')
        # Check appropriate boxes.
        self.checkDiameter.select()
        self.checkLength.select()
        self.checkArea.select()
        self.checkSumConnectedLength.select()
        self.checkAreaFraction.select()
                
        # Segment options.
        # Variables.
        self.varSegment = tk.StringVar()
        # Create widgets.
        self.radAccurate = tk.Radiobutton(self.frameSegment,
                text="Accurate", variable=self.varSegment, value="accurate",
                command=self.updateOptions)
        self.radFast = tk.Radiobutton(self.frameSegment,
                text="Fast", variable=self.varSegment, value="fast",
                command=self.updateOptions)
        self.radManual= tk.Radiobutton(self.frameSegment,
                text="Manual", variable=self.varSegment, value="manual",
                command=self.updateOptions)
        self.radFromBinary = tk.Radiobutton(self.frameSegment,
                text="From binary", variable=self.varSegment, value="binary",
                command=self.updateOptions)
        # Pack widgets.
        self.radAccurate.grid(row=0, column=0, sticky='w')
        self.radFast.grid(row=1, column=0, sticky='w')
        self.radManual.grid(row=2, column=0, sticky='w')
        self.radFromBinary.grid(row=3, column=0, sticky='w')
        # Check appropriate boxes.
        self.radAccurate.select()

        # Option options.
        # Profiles
        profiles = autoSelect.profiles()
        # Variables.
        self.varShowSteps = tk.BooleanVar()
        self.varOutputExcel = tk.BooleanVar()
        self.varSavePDF = tk.BooleanVar()
        self.varSaveMovie = tk.BooleanVar()
        self.varSaveBinary = tk.BooleanVar()
        self.varAutoParse = tk.BooleanVar()
        self.varProfile = tk.StringVar()
        self.varProfile.set(profiles[0])
        # Create widgets.
        self.checkShowSteps = tk.Checkbutton(self.frameOptions,
                text="Show steps", variable=self.varShowSteps)
        self.checkOutputExcel = tk.Checkbutton(self.frameOptions,
                text="Output to Excel", variable=self.varOutputExcel)
        self.checkSavePDF = tk.Checkbutton(self.frameOptions,
                text="Save PDF", variable=self.varSavePDF)
        self.checkSaveMovie = tk.Checkbutton(self.frameOptions,
                text="Save movie", variable=self.varSaveMovie)
        self.checkSaveBinary = tk.Checkbutton(self.frameOptions,
                text="Save binary", variable=self.varSaveBinary)
        self.checkAutoParse = tk.Checkbutton(self.frameOptions,
                text="Auto parse raw image", variable=self.varAutoParse,
                command=self.updateAuto)
        self.optionProfile = tk.OptionMenu(self.frameOptions, self.varProfile,
                *profiles)
        self.optionProfile.config(state=tk.DISABLED)

        # Pack widgets.
        self.checkShowSteps.grid(row=0, column=0, sticky='w')
        self.checkOutputExcel.grid(row=1, column=0, sticky='w')
        self.checkSavePDF.grid(row=2, column=0, sticky='w')
        #self.checkSaveMovie.grid(row=3, column=0, sticky='w')
        self.checkSaveBinary.grid(row=4, column=0, sticky='w')
        #self.checkAutoParse.grid(row=5, column=0, sticky='w')
        #self.optionProfile.grid(row=6, column=0, sticky='w', padx=15)
        
        # Check appropriate boxes.
        self.checkOutputExcel.select()

        self.createToolTips()

    def createToolTips(self):
        self.ttps = []
        this_dir, this_filename = os.path.split(__file__)
        widgets = [[self.radAccurate, 'seg_accurate.txt'],
                   [self.radFast, 'seg_fast.txt'],
                   [self.radManual, 'seg_manual.txt'],
                   [self.radFromBinary, 'seg_binary.txt'],
                   [self.checkShowSteps, 'opt_show_steps.txt'],
                   [self.checkOutputExcel, 'opt_output_excel.txt'],
                   [self.checkSavePDF, 'opt_save_pdf.txt'],
                   [self.checkSaveMovie, 'opt_save_movie.txt'],
                   [self.checkSaveBinary, 'opt_save_binary.txt'],
                   [self.checkDiameter, 'mes_diameter.txt'],
                   [self.checkLength, 'mes_length.txt'],
                   [self.checkArea, 'mes_area.txt'],
                   [self.checkSumConnectedLength, 'mes_connected_length.txt'],
                   [self.checkAreaFraction, 'mes_area_fraction.txt'],
                   ]

        for widget, txt in widgets:
            text_path = os.path.join(this_dir, 'ttps', txt)
            f = open(text_path)
            data = f.read()
            f.close()
            self.ttps.append(CreateToolTip(widget, data))

    def setOutputFolder(self):
        self.outFold = selectFolder()
        
    def saveAll(self):
        """
        Saves the results of each image analyzed this session to an excel file.
        """
        path = saveFile(ftype='xlsx')
        writer = pd.ExcelWriter(path)
        df = pd.DataFrame(self.saveAll)
        df.to_excel(writer, header=False, index=False)
        writer.save()
        
        #Format the excel file
        try:
            import openpyxl
            from openpyxl.styles import Alignment, Font, Border, Side
            #Load the workbook and worksheet
            wb = openpyxl.load_workbook(filename=path)
            ws = wb.get_sheet_by_name("Sheet1")
            cells = ['E1','H1','K1','N1','Q1','T1','W1','Z1']
            ws.merge_cells('E1:G1')
            ws.merge_cells('H1:J1')
            ws.merge_cells('K1:M1')
            ws.merge_cells('N1:P1')
            ws.merge_cells('Q1:S1')
            ws.merge_cells('T1:V1')
            ws.merge_cells('W1:Y1')
            ws.merge_cells('Z1:AB1')
            #Bold and center the headers
            ft = Font(bold=True)
            for cell in cells:
                ws[cell].alignment = Alignment(horizontal="center")
                ws[cell].font = ft
            #Add borders
            rows,_ = self.saveAll.shape
            for i in range(rows):
                for cell in cells:
                    c = cell[0]+str(i+1)
                    ws[c].border = Border(left=Side(style='thin'))

                
            
            wb.save(path)
            
        except ImportError:
            pass

    def updateOptions(self):
        """
        Updates the GUI options.  Deactivates some options in response to the
        selection of other options.
        """
        if self.varSegment.get() == "binary":
            self.checkSaveBinary.config(state=tk.DISABLED)
        else:
            self.checkSaveBinary.config(state=tk.NORMAL)

    def updateAuto(self):
        """
        Makes the profile widget visible.
        """
        if self.varAutoParse.get():
            self.optionProfile.config(state=tk.NORMAL)
        else:
            self.optionProfile.config(state=tk.DISABLED)

    def write(self, box, *strings, sep=' ', end='\n'):
        """
        Displays text in a text box with similar functionality as the print
        statement.

        Parameters
        ----------
        box : tk.Text or ScrolledText
            Text box that will be written to.

        *strings : str (any number of)
            Strings that will be written in the text box.
        sep : str, optional
            Text that will be inserted between *strings arguements.
        end : str, optional
            Text that will be inserted at the end.
        """

        # Prepare the text to write.
        output = ''
        for i in range(len(strings)-1):            
            output = output+str(strings[i])+sep
        output = output+str(strings[-1])
        output = output + end

        # Write the output to the appropraite text box.
        box.insert(tk.END, output)
        box.see(tk.END)

        self.update_idletasks()
        time.sleep(0.1)

    def prepare(self):
        """
        Allows the selection of multiple files if segment is not set to binary.
        Calls the calculate function for each selected file.
        """
        if self.varSegment.get() == "binary":
            self.calculate("")
        else:
            files = selectFile(multiple=True)
            for file in files:
                self.calculate(file)

    def about(self):
        """
        Displays the about the software window.
        """
        
        try:
            if self.tlAbout.winfo_exists():
                exists = True
            else:
                exists = False
        except AttributeError:
            exists = False

        if not exists:    
            self.tlAbout = tk.Toplevel(self)
            self.tlAbout.title('About AQUAMI')
            module_path, this_filename = os.path.split(__file__)
            self.tlAbout.iconbitmap(''.join((module_path, '/icon.ico')))
            self.buttonAboutClose = tk.Button(self.tlAbout, text='OK', width=8,
                                              command=self.aboutClose)
            self.frameAbout = tk.Frame(self.tlAbout)
            self.frameAbout.grid(row=0, column=0, padx=5, pady=5, ipadx=0, ipady=0)
            self.buttonAboutClose.grid(row=1, column=0, sticky='e', pady=5, padx=15)

            self.aText = tk.Text(self.frameAbout, background='SystemButtonFace',
                                 width=75, height=18, font="Times 11", bd=0,
                                 exportselection=False, cursor='arrow')
            self.aText.pack(anchor='w')

            self.aText.insert(tk.INSERT, "AQUAMI\n", )
            self.aText.tag_add('title', '1.0', '1.20')
            self.aText.tag_config('title', font='Times 16 bold')
            self.aText.insert(tk.END, ''.join((
                'Version: ', __version__, '\n', 
                'Author: Joshua Stuckner\n',
                'Contact: stuckner@vt.edu\n',
                'Liscence: MIT\n\n',
                'This software was developed by a graduate student at Virginia Tech and released in\n',
                'the hopes that it will be useful. This software may be freely modified and/or distributed. \n',
                'The author plans to support this software until at least Dec. 2018. Please contact the\n',
                'author to report any issues or bugs, or to request additional features.  Highly demanded\n',
                'features may be added to the software.\n\n',
                'If this software is useful to your research, please consider citing:\n',
                'J. A. Stuckner et al., "AQUAMI: An Open Source Python Package and GUI for the\n',
                'Automatic Quantitative Analysis of Morphologically Complex Multiphase Materials,"\n',
                'Computational Materials Science, vol. 139, pp. 320-329, Nov. 2017\n',
                'DOI: 10.1016/j.commatsci.2017.08.012'
                )))

            self.aText.tag_add('citing', '13.0', '13.100')
            self.aText.tag_config('citing', font='Times 11 bold')
            self.aText.tag_add('etal', '14.15', '14.21')
            self.aText.tag_config('etal', font='Times 11 italic')
            self.aText.tag_add('journal', '16.0', '16.31')
            self.aText.tag_config('journal', font='Times 11 italic')
            self.aText.tag_add('doi', '17.0', '17.111')
            self.aText.tag_config('doi', foreground='blue', underline=1)
            self.aText.tag_bind('doi', '<Button-1>', self.hyperlink_click)
            self.aText.tag_bind("doi", "<Enter>", self.hyper_enter)
            self.aText.tag_bind("doi", "<Leave>", self.hyper_leave)
            
            self.aText.config(state=tk.DISABLED)

    def hyperlink_click(self, event):
        url = r'https://doi.org/10.1016/j.commatsci.2017.08.012'
        if sys.platform=='win32':
            os.startfile(url)
        elif sys.platform=='darwin':
            subprocess.Popen(['open', url])
        else:
            try:
                subprocess.Popen(['xdg-open', url])
            except OSError:
                print(''.join(('Please open a browser on: ', url)))

    def hyper_enter(self, event):
            self.aText.config(cursor="hand2")

    def hyper_leave(self, event):
        self.aText.config(cursor="arrow")

    def aboutClose(self):
        """
        Closes the about the software window
        """
        self.tlAbout.destroy()
        

    def calculate(self, fullFile):
        """
        Performs measurements based on the options.

        Parameters
        ----------
        fullFile : str
            The path to the image to operate on.
        """

        # Initialize variables
        pixelSize = 0
        areaFracBright = 0
        estSize= 0
        diameter_average_bright = 0
        diameter_SD_bright = 0
        diameter_num_bright = 0
        diameter_average_dark = 0
        diameter_SD_dark = 0
        diameter_num_dark = 0
        length_average_bright = 0
        length_SD_bright = 0
        length_num_bright = 0
        length_average_dark = 0
        length_SD_dark = 0
        length_num_dark = 0
        area_average_bright = 0
        area_SD_bright = 0
        area_num_bright = 0
        area_average_dark = 0
        area_SD_dark = 0
        area_num_dark = 0
        sumLength_average_bright = 0
        sumLength_SD_bright = 0
        sumLength_num_bright = 0
        sumLength_average_dark = 0
        sumLength_SD_dark = 0
        sumLength_num_dark = 0
        
        # Rename text boxes and tkinter variables for easier reference
        results = self.textResults
        status = self.textStatus
        showSteps = self.varShowSteps.get()
        outputExcel = self.varOutputExcel.get()
        savePDF = self.varSavePDF.get()

        # Handle strange GUI input.
        if self.varSegment.get() == "binary":
            self.varSaveBinary.set(False)

        # Select the image to operate on.
        if self.varSegment.get() == "binary":
            brightFile = selectFile(title="Select the bright phase mask")
            darkFile = selectFile(title="Select the bright phase mask")
            folder = '/'.join(brightFile.split('/')[:-1])
            fname = brightFile.split('/')[-1].split(' -')[:-1][0]
            ftype = brightFile.split('.')[-1]
            fnametype = fname + '.' + ftype
            pixelString = brightFile.split('--')[1]            
            pixelSize = float(pixelString)
            fullFile = brightFile

        else:
            #fullFile = selectFile()
            # Figure out file name and type.
            folder = '/'.join(fullFile.split('/')[:-1])
            fname = fullFile.split('/')[-1].split('.')[:-1][0]
            ftype = fullFile.split('.')[-1]
            fnametype = fname + '.' + ftype

        # Status update.
        if self.varSegment.get() == "binary":
            self.write(status, "Operating on", fnametype, "(binary)")
            self.write(results, "Operating on", fnametype, "(binary)")
        else:
            self.write(status, "Operating on", fnametype)
            self.write(results, "Operating on", fnametype)

        # Load image.
        img_raw = inout.load(fullFile)

        # Prepare for output of data to excel
        if outputExcel:
            if self.outFold == None:
                excelPath = folder+'/'+fname+' - All Measurements.xlsx'
                writer = pd.ExcelWriter(excelPath)
            else:
                excelPath = self.outFold+'/'+fname+' - All Measurements.xlsx'
                writer = pd.ExcelWriter(excelPath)

        # Prepare PDF file.
        pdf = None
        if savePDF:
            if self.outFold == None:
                pdf = PdfPages(folder+'/'+fname+' - Steps.pdf')
            else:
                pdf = PdfPages(self.outFold+'/'+fname+' - Steps.pdf')

        # Show steps.
        if showSteps:
            display.showFull(img_raw, title="Raw Image.")

        if savePDF:
            inout.pdfSaveImage(pdf, img_raw, title="Raw Image.")

        # Select the image data and scale bar.
        if self.varSegment.get() != "binary":

            if self.varAutoParse.get():
                profile = self.varProfile.get()
                img, pixelSize = autoSelect.autoDetect(img_raw, profile=profile)
            else:
                img, scale = manualSelectImage(img_raw)
                
                # Ensure proper dtype.
                img = inout.uint8(img) 
                scale = inout.uint8(scale)

                # Get the pixel size.
                pixelSize = measure.manualPixelSize(scale)
                self.write(results, "The pixel size is %.3f nm" %(pixelSize))
            
            # Get a rough estimated of the ligament diameter.
            bright, dark = segment.roughSegment(img)
            estSizeD = measure.estimateDiameter(dark)
            estSizeB = measure.estimateDiameter(bright)
            estSizeB_rough = copy.copy(estSizeB)
            estSizeD_rough = copy.copy(estSizeD)
            # Get the average estimated size with a weight towards the thinner
            estSize = (min(estSizeD, estSizeB) + estSizeD + estSizeB) / 3
            # Result update                
            self.write(results, "Weighted average estimated diameter: ",
                       round(estSize,3)," pixels (", round(estSize*pixelSize,3),
                       " nm).", sep='')
            if savePDF:
                inout.pdfSaveImage(pdf, img, title="Selected image data",
                                   cmap=plt.cm.gray)
        else:
            img = img_raw.copy()
            img = inout.uint8(img)

        # Segment the image.
        # Status update.
        if self.varSegment.get() == "accurate":
            self.write(status, "Segmenting image. This may take a while,",
                       "especially at higher resolutions.")
        else:
            self.write(status, "Segmenting image.")
        # Segment based on options.    
        if self.varSegment.get() == "fast":
            #already did this when estimating size
            pass
        elif self.varSegment.get() == "manual":
            bright, dark = segment.manualSegment(img)
        elif self.varSegment.get() == "accurate":
            bright, dark = segment.segment(img, estSize)
        elif self.varSegment.get() == "binary":
            bright = imread(brightFile)
            dark = imread(darkFile)
            estSizeD = measure.estimateDiameter(dark)
            estSizeB = measure.estimateDiameter(bright)
            estSizeB_rough = copy.copy(estSizeB)
            estSizeD_rough = copy.copy(estSizeD)
            # Get the average estimated size with a weight towards the thinner
            estSize = (min(estSizeD, estSizeB) + estSizeD + estSizeB) / 3
            # Result update                
            self.write(results, "Weighted average estimated diameter: ",
                       round(estSize,3)," pixels (", round(estSize*pixelSize,3),
                       " nm).", sep='')
            bright = inout.uint8(bright)
            dark = inout.uint8(dark)

        if showSteps and self.varSegment.get() != "binary":
            display.overlayMask(img, bright, title="Bright phase mask.",
                                animate=True)
            display.overlayMask(img, dark, title="Dark phase mask.",
                                animate=True)

        if savePDF and self.varSegment.get() != "binary":
            inout.pdfSaveOverlay(pdf, img, bright, title="Bright phase mask.")
            inout.pdfSaveOverlay(pdf, img, dark, title="Dark phase mask.")

        if self.varSaveMovie.get():
            stack = []
            for i in range(5):
                stack.append(gray2rgb(img))
                stack.append(display.overlayMask(img,bright,
                                                 return_overlay=True))
            inout.save_movie(stack, ''.join((folder, '//', fname,
                                             ' - segment movie.mp4')))

        if self.varSaveBinary.get():
            inout.saveBinary(bright,
                folder+'/'+fname+' - Bright Segment--'+str(pixelSize)+'--.tif')
            inout.saveBinary(dark,
                folder+'/'+fname+' - Dark Segment--'+str(pixelSize)+'--.tif')

        
              
        # Calculate area fraction.
        if self.varAreaFraction.get():
            # Status update.
            self.write(status, "Calculating area fraction.")
            areaFracBright = np.count_nonzero(bright) /        \
                    (np.count_nonzero(bright) + np.count_nonzero(dark))
            # Results update.
            self.write(results, "Bright phase area fraction: %.3f" \
                       %(areaFracBright))

        # Refine the estimated diameters.
        estSizeB, _ = measure.diameter(bright, estSize)
        estSizeD, _ = measure.diameter(dark, estSize)

                           
        # Calculate the diameter.
        if self.varDiameter.get():
            # Status update.
            self.write(status, "Calculating ligament diameter.")
            # Calculate.
            diameter_average_bright, diameter_SD_bright, diameter_all_bright =\
                            measure.diameter(bright, estSizeB,
                                             showSteps=showSteps,
                                             returnAll=True,
                                             pdf=pdf)
            diameter_num_bright = len(diameter_all_bright)
            diameter_average_dark, diameter_SD_dark, diameter_all_dark =\
                            measure.diameter(dark, estSizeD,
                                             showSteps=showSteps,
                                             returnAll=True,
                                             pdf=pdf)
            diameter_num_dark = len(diameter_all_dark)
            # Update estimated ligament sizes.
            estSizeB = diameter_average_bright
            estSizeD = diameter_average_dark
            # Results update.
            self.write(results, "Bright phase diameter:",
                       round(diameter_average_bright,2), "±",
                       round(diameter_SD_bright), "pixels.")
            self.write(results, "Dark phase diameter:",
                       round(diameter_average_dark,2), "±",
                       round(diameter_SD_dark), "pixels.")

            if outputExcel:
                df = pd.DataFrame(diameter_all_bright)
                df.to_excel(writer, sheet_name="Bright phase diameter data")
                df = pd.DataFrame(diameter_all_dark)
                df.to_excel(writer, sheet_name="Dark phase diameter data") 
                
        # Fix estSizeB and estSizeD if there is a problem
        estSizeB = estSizeB_rough if np.isnan(estSizeB) else estSizeB      
        estSizeD = estSizeD_rough if np.isnan(estSizeD) else estSizeD         
            
        # Calculate the ligament length.
        if self.varLength.get():
            # Status update.
            self.write(status, "Calculating ligament length.")
            length_average_bright, length_SD_bright, length_all_bright = \
                            measure.length(bright, estSizeB,
                                           showSteps=showSteps,
                                           returnAll=True,
                                           pdf=pdf)
            try:
                length_num_bright = len(length_all_bright)
            except TypeError:
                length_num_bright = 0
            length_average_dark, length_SD_dark, length_all_dark = \
                            measure.length(dark, estSizeD,
                                           showSteps=showSteps,
                                           returnAll=True,
                                           pdf=pdf)
            try:
                length_num_dark = len(length_all_dark)
            except TypeError:
                length_num_dark = 0
            # Results update.
            if length_average_bright != 0:
                self.write(results, "Bright phase ligament length:",
                           round(length_average_bright,2), "±",
                           round(length_SD_bright), "pixels.")
            if length_average_dark != 0:
                self.write(results, "Dark phase ligament length:",
                           round(length_average_dark,2), "±",
                           round(length_SD_dark), "pixels.")

            if outputExcel:
                if length_average_bright != 0:
                    df = pd.DataFrame(length_all_bright)
                    df.to_excel(writer, sheet_name="Bright phase length data")
                if length_average_dark != 0:
                    df = pd.DataFrame(length_all_dark)
                    df.to_excel(writer, sheet_name="Dark phase length data") 

        # Calculate average object area.
        if self.varArea.get():
            # Status update.
            self.write(status, "Calculating average object area.")
            area_average_bright, area_SD_bright, area_all_bright = \
                            measure.area(bright, estSizeB,
                                           showSteps=showSteps,
                                           returnAll=True)
            try:
                area_num_bright = len(area_all_bright)
            except TypeError:
                area_num_bright = 0
            area_average_dark, area_SD_dark, area_all_dark = \
                            measure.area(dark, estSizeD,
                                           showSteps=showSteps,
                                           returnAll=True)
            try:
                area_num_dark = len(area_all_dark)
            except TypeError:
                area_num_dark = 0
            # Results update.
            if area_average_bright != 0:
                self.write(results, "Bright phase average object area:",
                           round(area_average_bright,2), "±",
                           round(area_SD_bright), "pixels^2.")
            if area_average_dark != 0:
                self.write(results, "Dark phase average object area:",
                           round(area_average_dark,2), "±",
                           round(area_SD_dark), "pixels^2.")

            if outputExcel:
                if area_average_bright != 0:
                    df = pd.DataFrame(area_all_bright)
                    df.to_excel(writer, sheet_name="Bright phase area data")
                if area_average_dark != 0:
                    df = pd.DataFrame(area_all_dark)
                    df.to_excel(writer, sheet_name="Dark phase area data") 

        # Calculate the sum connected length of each object.
        if self.varSumConnectedLength.get():
            # Status update.
            self.write(status, "Calculating sum connected ligament length.")
            sumLength_average_bright, sumLength_SD_bright, \
                                      sumLength_all_bright = \
                            measure.connectedLength(bright, estSizeB,
                                        showSteps=showSteps,
                                        returnAll=True,
                                        pdf=pdf)
            try:
                sumLength_num_bright = len(sumLength_all_bright)
            except TypeError:
                sumLength_num_bright = 0
            sumLength_average_dark, sumLength_SD_dark, sumLength_all_dark = \
                            measure.connectedLength(dark, estSizeD,
                                        showSteps=showSteps,
                                        returnAll=True,
                                        pdf=pdf)
            try:
                sumLength_num_dark = len(sumLength_all_dark)
            except TypeError:
                sumLength_num_dark = 0
            # Results update.
            if sumLength_average_bright != 0:
                self.write(results, "Bright phase sum connected length:",
                           round(sumLength_average_bright,2), "±",
                           round(sumLength_SD_bright), "pixels.")
            if sumLength_average_dark != 0:
                self.write(results, "Dark phase sum connected length:",
                           round(sumLength_average_dark,2), "±",
                           round(sumLength_SD_dark), "pixels.")

            if outputExcel:
                if sumLength_average_bright != 0:
                    df = pd.DataFrame(sumLength_all_bright)
                    df.to_excel(writer,
                        sheet_name="Bright phase sum length data")
                if sumLength_average_dark != 0:
                    df = pd.DataFrame(sumLength_all_dark)
                    df.to_excel(writer,
                        sheet_name="Dark phase sum length data") 

        # Save excel file
        if outputExcel:
            writer.save()

        # Save the pdf.
        if savePDF:
            plt.title('Summary')
            plt.axis('off')
            fs = 10
            yloc = 0.95
            xloc = 0.01
            space = 0.05
            out = 'Image: ' + fnametype
            plt.text(xloc, yloc, out, fontsize=fs)
            yloc -= space
            out = 'Pixel size: ' + str(pixelSize)
            plt.text(xloc, yloc, out, fontsize=fs)
            yloc -= space
            out = 'Bright phase area fraction: ' + str(round(areaFracBright,3))
            plt.text(xloc, yloc, out, fontsize=fs)
            yloc -= space
            out = 'Rough diameter estimate: ' + str(round(estSize,3))+' pixels.'
            plt.text(xloc, yloc, out, fontsize=fs)
            yloc -= space*2

            out = 'Bright phase:'
            plt.text(xloc, yloc, out, fontsize=fs, fontweight='bold')
            yloc -= space
            if diameter_average_bright != 0:
                out = ('Ligament diameter: ' +
                       str(round(diameter_average_bright*pixelSize,3)) +
                       ' ± ' + str(round(diameter_SD_bright*pixelSize,3)) +
                       ' nm')
                plt.text(xloc, yloc, out, fontsize=fs)
                yloc -= space
            if length_average_bright != 0:
                out = ('Ligament length: ' +
                       str(round(length_average_bright*pixelSize,3)) +
                       ' ± ' + str(round(length_SD_bright*pixelSize,3)) +
                       ' nm')
                plt.text(xloc, yloc, out, fontsize=fs)
                yloc -= space
            if area_average_bright != 0:
                out = ('Object area: ' +
                       str(round(area_average_bright*pixelSize**2,3)) +
                       ' ± ' + str(round(area_SD_bright*pixelSize**2,3)) +
                       r' nm$^2$')
                plt.text(xloc, yloc, out, fontsize=fs)
                yloc -= space
            if sumLength_average_bright != 0:
                out = ('Sum connected length: ' +
                       str(round(sumLength_average_bright*pixelSize,3)) +
                       ' ± ' + str(round(sumLength_SD_bright*pixelSize,3)) +
                       ' nm')
                plt.text(xloc, yloc, out, fontsize=fs)
                yloc -= space
            yloc -= space

            out = 'Dark phase:'
            plt.text(xloc, yloc, out, fontsize=fs, fontweight='bold')
            yloc -= space
            if diameter_average_dark != 0:
                out = ('Ligament diameter: ' +
                       str(round(diameter_average_dark*pixelSize,3)) +
                       ' ± ' + str(round(diameter_SD_dark*pixelSize,3)) +
                       ' nm')
                plt.text(xloc, yloc, out, fontsize=fs)
                yloc -= space
            if length_average_dark != 0:
                out = ('Ligament length: ' +
                       str(round(length_average_dark*pixelSize,3)) +
                       ' ± ' + str(round(length_SD_dark*pixelSize,3)) +
                       ' nm')
                plt.text(xloc, yloc, out, fontsize=fs)
                yloc -= space
            if area_average_dark != 0:
                out = ('Object area: ' +
                       str(round(area_average_dark*pixelSize**2,3)) +
                       ' ± ' + str(round(area_SD_dark*pixelSize**2,3)) +
                       r' nm$^2$')
                plt.text(xloc, yloc, out, fontsize=fs)
                yloc -= space
            if sumLength_average_dark != 0:
                out = ('Sum connected length: ' +
                       str(round(sumLength_average_dark*pixelSize,3)) +
                       ' ± ' + str(round(sumLength_SD_dark*pixelSize,3)) +
                       ' nm')
                plt.text(xloc, yloc, out, fontsize=fs)
                yloc -= space
            

            pdf.savefig()
            plt.close()
            pdf.close()

        # Append data to the saveAll
        newdata = np.array(
                [[fnametype,
                  round(pixelSize,2),
                  round(areaFracBright,3),
                  round(estSize*pixelSize,3),
                  round(diameter_average_bright*pixelSize,3),
                  round(diameter_SD_bright*pixelSize,3),
                  diameter_num_bright,
                  round(diameter_average_dark*pixelSize,3),
                  round(diameter_SD_dark*pixelSize,3),
                  diameter_num_dark,
                  round(length_average_bright*pixelSize,3),
                  round(length_SD_bright*pixelSize,3),
                  length_num_bright,
                  round(length_average_dark*pixelSize,3),
                  round(length_SD_dark*pixelSize,3),
                  length_num_dark,
                  round(area_average_bright*pixelSize**2,3),
                  round(area_SD_bright*pixelSize**2,3),
                  area_num_bright,
                  round(area_average_dark*pixelSize**2,3),
                  round(area_SD_dark*pixelSize**2,3),
                  area_num_dark,
                  round(sumLength_average_bright*pixelSize,3),
                  round(sumLength_SD_bright*pixelSize,3),
                  sumLength_num_bright,
                  round(sumLength_average_dark*pixelSize,3),
                  round(sumLength_SD_dark*pixelSize,3),
                  sumLength_num_dark
                  ]])
        

        
        self.saveAll = np.append(self.saveAll, newdata, axis=0)

        # Add new line to status and results
        self.write(results, '')
        self.write(status, '')
        
if __name__ == "__main__":
    myapp = Aquami_Gui()
    myapp.mainloop()

def run():
    myapp = Aquami_Gui()
    myapp.mainloop()




#!/usr/bin/env python3
"""
This module contains functions to perform measurements on BNMs. Includes
functions for skeletonization, distance transforms, and measuring ligament
length and width.
"""

import warnings

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from skimage.morphology import (skeletonize, remove_small_objects, disk,
                                binary_dilation, binary_opening)
from skimage.measure import label
from scipy.signal import convolve2d
from scipy.stats import lognorm, norm
from scipy.ndimage.morphology import distance_transform_edt
import tkinter as tk

from aquami import display
from aquami import gui
from aquami import inout
from aquami import segment

__author__ = "Joshua Stuckner"

def isConnected(mask):
    """
    Determines whether the masked phase is fully connected (returns True) or
    seperated (returns False)

    Parameters
    ----------
    mask : 2D array of bool
        Mask showing the area of one phase.

    Returns
    -------
    bool
        True if mask is mostly connected.  False otherwise.
    """

    # Label each blob.  0 = background, 1 = blobs touching edge.
    lig_labels = label(mask, background=0)

    # Find the area of each connected section.
    areas = np.bincount(lig_labels.flatten()) #count the pixels
    areas[1:][::-1].sort() #sort in descending order (ignoring the background).

    # If connected area touching the edge is larger than the sum of all other
    # areas then this phase is basically fully connected.
    # Use the appropriate skeletonizing function.
    if areas[1] > np.sum(areas[2:]):
        return True
    else:
        return False
    
def skelConnected(mask, diam, removeEdge=False, times=3):
    """
    Returns the skeletal backbone of the ligaments or pores which are one pixel
    thick lines sharing the same connectivity as the passed binary mask.

    Parameters
    ----------
    mask : 2D array of bool
        Mask showing the area of one phase. Must be continuously connected.
    diam : float64
        The estimated or measured average diameter of the ligaments in the mask.
    removeEdge : bool, optional
        If true, the returned skel does not include ligaments that are too close
        to the edge.

    Returns
    -------
    skel : 2D array of bool
        Skeletal backbone of the mask.  One pixel thick lines sharing the same
        connectivity as the mask.
    """

    # Get the shape of the mask.
    rows, cols = mask.shape

    # Find the initial skeletal backbone.
    skel = skeletonize(mask>0)
    
    # Remove  terminal edges.
    
    for i in range(times):
        # Find number of 1st and 2nd neighboring pixels
        neighbors = convolve2d(skel, np.ones((3,3)), mode='same')
        neighbors = neighbors * skel
        neighbors2 = convolve2d(skel, np.ones((5,5)), mode='same')
        neighbors2 = neighbors2 * skel
    
        # Remove nodes and label each ligament section.
        nodes = neighbors > 3
        skel = skel - nodes
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            skel = remove_small_objects(skel, 1, connectivity=1)
        labels = label(skel, connectivity=2, background=0)

        # Find terminal sections.
        terminal_labels = labels[neighbors==2] #These are definitly terminal
        # When there is a 1 pixel fork at the end, we need to look at second
        # nearest neighbors as well.
        #neighbors[neighbors2>5] = 10
        #terminal_labels2 = labels[neighbors<5]
        #terminal_labels = np.append(terminal_labels, terminal_labels2)
        terminal_sections = np.unique(terminal_labels)
        just_terminal = np.zeros(skel.shape, dtype=bool)
        for lab in terminal_sections:
            just_terminal[labels==lab] = 1

        # Remove terminal sections. 
        skel = skel - just_terminal
        
        # Put the nodes back in.
        skel = binary_dilation(skel, selem=disk(3))    
        skel = skeletonize(skel)

    # Remove sections that touch the edge of image.
    if removeEdge:
        try:
            edge = round(int(diam))
        except ValueError:
            edge = 0
        skel = skel - nodes
        labels = label(skel, connectivity=2, background=0)
        elabels = np.copy(labels)
        elabels[edge:rows-edge-1,edge:cols-edge-1] = 0
        edge_labels = np.unique(elabels)
        for lab in edge_labels:
            skel[labels==lab] = 0
            
        # Put the nodes back in.
        skel = skel+nodes    

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            skel = remove_small_objects(skel, 8, connectivity=2)

    return skel

def skelBlobs(mask, diam):
    """
    Returns the skeletal backbone of the ligaments or pores which are one pixel
    thick lines sharing the same connectivity as the passed binary mask.

    Parameters
    ----------
    mask : 2D array of bool
        Mask showing the area of one phase. Must be seperated blobs.
    diam : float64
        The estimated or measured average diameter of the ligaments in the mask.

    Returns
    -------
    skel : 2D array of bool
        Skeletal backbone of the mask.  One pixel thick lines sharing the same
        connectivity as the mask.
    """

    # Don't operate on original mask.
    im = np.copy(mask)

    # Smooth edges to reduce spurious results.
    im = binary_opening(im>0, selem=disk(1))

    # Get initial skeletonized image.
    skel = skeletonize(im>0)

    return skel

def skeleton(mask, diam, removeEdge=False, times=3):
    """
    Returns the skeletal backbone of the ligaments or pores which are one pixel
    thick lines sharing the same connectivity as the passed binary mask.

    Parameters
    ----------
    mask : 2D array of bool
        Mask showing the area of one phase.
    diam : float64
        The estimated or measured average diameter of the ligaments in the mask.
    removeEdge : bool, optional
        If true, the returned skel does not include ligaments that are too close
        to the edge.
        
    Returns
    -------
    skel : 2D array of bool
        Skeletal backbone of the mask.  One pixel thick lines sharing the same
        connectivity as the mask.
    """

    if isConnected(mask):
        return skelConnected(mask, diam, removeEdge=removeEdge, times=times)
    else:
        return skelBlobs(mask, diam)

def distanceTransform(mask):
    """
    Returns the euclidean distance each white pixel is to the nearest black
    pixel.  Just calls scipy.ndimage.morphology.distance_transform_edt.

    Parameters
    ----------
    mask : 2D array of bool
        Mask showing the area of one phase.

    Returns
    -------
    dist : ndarray
        Size in nm of each pixel
    """

    return distance_transform_edt(mask)
    
def manualPixelSize(scale):
    '''
    Returns the size of each pixel in nm.  Must manually select the scale bar
    and input the size.

    Parameters
    ----------
    scale : ndarray
        Image containing the scale bar

    Returns
    -------
    pixelSize : float64
        Size in nm of each pixel
    '''

    plt.imshow(scale, cmap=plt.cm.gray, interpolation=None)
    plt.title("Draw a rectangle the same width as the scale bar and close the"+\
            " figure.")
    figManager = plt.get_current_fig_manager()
    figManager.window.showMaximized()
    a = gui.selectRect()
    plt.axis('off')
    plt.show()
    barSize = a.x1-a.x0
    intScaleNum = 0
    root = tk.Tk()
    pixapp = gui.inputScale(root, scale=scale)
    intScaleNum = pixapp.result
    root.destroy()
    try:
        pixelSize = round(intScaleNum/barSize, 2) # More decimals are arbitrary.
    except TypeError:
        pixelSize = 0
    return pixelSize

def estimateDiameter(mask):
    '''
    Returns a rough estimate of the ligament size in order to set parameters
    for more refined thresholding and size calculations.

    Parameters
    ----------
    mask : ndarray
        Binary image file containing thresholded data

    debug : boolean, option
        Set true to display the process

    Returns
    -------
    estSize: float64
        Rough estimate of ligament average radius
    '''
    
    rows, cols = mask.shape
    sizes = np.zeros((2048))
    count = 0 #keeps track of number of pixels

    # Horizontal lines
    for i in range(rows):
        count = 0
        for j in range(cols):
            pixel = mask[i,j]
            if pixel:
                count += 1
            else:
                if count != 0:
                    sizes[count] += 1
                count = 0
            if j == cols - 1 and count != 0:
                sizes[count] += 1

    # Vertical lines
    for j in range(cols):
        count = 0
        for i in range(rows):
            pixel = mask[i,j]
            if pixel:
                count += 1
            else:
                if count != 0:
                    sizes[count] += 1
                count = 0
            if j == cols - 1 and count != 0:
                sizes[count] += 1

    #Calculate average diameter
    sumation = 0
    for i in range(5, len(sizes)):
        sumation += i*sizes[i]
    estSize = sumation / np.sum(sizes)

    return estSize

def diameter(mask, estSize, returnAll=False, showSteps=False, pdf=None):
    """
    Calculates the average ligament size and standard deviation of the passed
    mask.

    Parameters
    ----------
    mask : 2D array of bool
        Mask showing the area of one phase.
    estSize : float64
        The estimated diameter.  Used to set some parameters.
    returnAll : bool, optional
        If set to true, a list of all the measurements will be returned.
    showSteps : bool, optional
        If set to true, figures will be displayed that shows the results of
        intermediate steps when calculating the diameter.
    pdf : matplotlib.backends.backend_pdf.PdfPages, optional
        If provided, information will be output to the pdf file.
        
    Returns
    -------
    average : float64
        Average ligament diameter.  (Assumes gaussian distribution.)
    SD : float64
        Standard deviation of ligament diameters.
    allMeasurements: 1D array of float64, optional
        A list of all measured ligament diameters.
    """

    rows, cols = mask.shape
    
    # Adjustable parameters.
    # These have been highly tested and changing these may throw off results.
    try:
        ignore_edge_dist = int(estSize * 3) #Ignores results this close to the edge.
    except ValueError:
        #This only happens if estSize is NaN
        ignore_edge_dist = 70 
    
    # make sure the ignored edge isn't too big
    narrow = min(rows,cols)    
    if ignore_edge_dist > narrow / 3:
        ignore_edge_dist = int(narrow/6)
        
    thresh = estSize / 20 # Diameter's smaller than this are ignored as noise.
    
    # Get the distance transform and the skelatal backbone.
    dist = distanceTransform(mask)
    dist = dist * 2 # Convert radius to diameter.
    skel = skeleton(mask, estSize)
    

    # Need to do more processing on the skel if non connected phase.
    if not isConnected(mask):
        # Remove end-of-line pixels from the skel.
        pixels_to_remove = int(estSize / 2)
        for i in range(pixels_to_remove):
            keep = np.ones(skel.shape)
            terminal_ends = convolve2d(skel, np.ones((3,3)), mode='same')
            keep[terminal_ends < 3] = 0
            skel *= keep > 0

    # Multiply them together to get the diameter measurements along the backbone
    # of each ligament.
    dist *= skel

    if showSteps:
        dialate = False if cols < 1500 else True #Thicker result at high res.
        display.showSkel(dist, mask, dialate=dialate,
                         title="Ligament diameters along the backbone."
                         )
    if pdf is not None:
        dialate = False if cols < 1500 else True #Thicker result at high res.
        inout.pdfSaveSkel(pdf, dist, mask, dialate=dialate,
                         title="Ligament diameters along the backbone."
                         )
    # Remove the edges of the distance map.
    dist = dist[ignore_edge_dist:-ignore_edge_dist,
                ignore_edge_dist:-ignore_edge_dist]

    # Prepare list of all measurements.
    allMeasurements = np.sort(dist[dist>0].flatten())
    
    # Ignore small measurements as noise.
    dist = dist[dist>thresh]

    if showSteps:
        display.showHist(dist, gauss=True, log=True, title='Ligament Diameters',
                         xlabel='Ligament diameter [pixels]')

    if pdf is not None:
        inout.pdfSaveHist(pdf, dist, gauss=True, log=True,
                          title='Ligament Diameters',
                          xlabel='Ligament diameter [pixels]')
        
    average = np.average(dist)
    SD = np.std(dist)
    
    if returnAll:
        return average, SD, allMeasurements
    else:
        return average, SD

def length(mask, estSize, returnAll=False, showSteps=False, pdf=None):
    """
    Calculates the average between node ligament length and standard deviation
    of the passed mask.  Only meaningful on the fully connected phase.

    Parameters
    ----------
    mask : 2D array of bool
        Mask showing the area of one phase. 
    estSize : float64
        The estimated diameter.  Used to set some parameters.
    returnAll : bool, optional
        If set to true, a list of all the measurements will be returned.
    showSteps : bool, optional
        If set to true, figures will be displayed that shows the results of
        intermediate steps when calculating the diameter.
    pdf : matplotlib.backends.backend_pdf.PdfPages, optional
        If provided, information will be output to the pdf file.
        
    Returns
    -------
    logavg : float64
        Average ligament length.  (Assumes lognormal distribution.)
    logstd : float64
        Standard deviation of ligament diameters.
    allMeasurements: 1D array of float64, optional
        A list of all measured ligament lengths.
    """

    # Debug mode
    debug = False
    if debug:
        showSteps = True

    # Adjustable parameters.
    # These have been highly tested and changing these may throw off results.
    small_lig = 0.6*estSize # Ligaments smaller than this are within a node.
    filter_out_mult = 0.4 # This * estSize is ignored in final calculation.
    
    # Return 0 if the mask is not fully connected.
    if not isConnected(mask):
        if returnAll:
            return 0,0,0
        else:
            return 0,0
        
    rows, cols = mask.shape
    skel = skeleton(mask, estSize, removeEdge=True, times=3)
    
    if debug:
        display.showSkel(skel, mask)

    # get ligament and node labels
    nodes = np.copy(skel)
    ligaments = np.copy(skel)
    neighbors = convolve2d(skel, np.ones((3,3)), mode='same')
    neighbors = neighbors * skel
    nodes[neighbors < 4] = 0
    ligaments[neighbors > 3] = 0
    ligaments = label(ligaments, background=0)
    nodes = binary_dilation(nodes, selem=disk(3))
    nodes = label(nodes, background=0)    
    
    # get a list of ligaments connected to each node
    node_to_lig = []  # node_to_lig[n] is an array of each ligament label that is connected to node n
    unodes = np.unique(nodes)
    for n in unodes:
        node_to_lig.append(np.unique(ligaments[nodes==n]))

    # get a list of nodes connected to each ligament
    lig_to_node = []  # lig_to_node[l] is an array of each node label that is connected to ligament l
    uligs = np.unique(ligaments)
    for l in uligs:
        lig_to_node.append(np.unique(nodes[ligaments==l]))
        
    # Get the length of each ligament between nodes.
    lengths = np.bincount(ligaments.flatten())

    # Add ligaments that are within a single node to connected ligaments.
    small = int(round(small_lig))
    too_small = []
    for l in uligs:
        if lengths[l] <= small:
            too_small.append(l) #keep track of ligaments that are too small
            for connected in lig_to_node[l]:
                if connected > 0 and connected != l:
                    # add half the small ligament to the connected ligaments.
                    lengths[connected] += int(round(lengths[l]/2,0))
                    
    # Set to True to show which ligaments are considered to be within a
    # single node. 
    if showSteps:       
        ligaments_small = ligaments > 0
        ligaments_small = ligaments_small.astype('uint8')
        ligaments_small[ligaments_small==1] = 2
        for i in too_small:
            ligaments_small[ligaments==i] = 1
        display.showSkel(ligaments_small, mask, dialate=False,
                title=("Green = within node ligaments, " \
                       "White = between node ligaments"))


    # filter out background and extra small lengths
    lengths[0] = 0
    allMeasurements = lengths[lengths>0]
    lengths=lengths[lengths>estSize*filter_out_mult]

    if len(lengths) == 0:
        if returnAll:
            return 0,0,0
        else:
            return 0,0

   
    # Get a lognormal fit.
    fit = lognorm.fit(lengths.flatten(), floc=0)
    pdf_fitted = lognorm.pdf(lengths.flatten(),
                             fit[0], loc=fit[1], scale=fit[2])

    # Get gaussian fit.
    gfit = norm.fit(lengths.flatten())

    
    # Get average and standard deviation for lognormal fit.
    logaverage = lognorm.mean(fit[0], loc=fit[1], scale=fit[2])
    logstd = lognorm.std(fit[0], loc=fit[1], scale=fit[2])


    # Get average and standard deviation for normal fit.
    average = norm.mean(gfit[0], gfit[1])
    std = norm.std(gfit[0], gfit[1])
    numbMeasured = lengths.size

    if showSteps:
        display.showSkel(ligaments, mask, dialate=False)

    if showSteps:
        display.showHist(lengths, gauss=True, log=True,
                         title='Ligament Lengths',
                         xlabel='Ligament lengths [pixels]')

    if pdf is not None:
        inout.pdfSaveSkel(pdf, ligaments, mask, dialate=True)
        inout.pdfSaveHist(pdf, lengths, gauss=True, log=True,
                       title='Ligament Lengths',
                       xlabel='Ligament lengths [pixels]')
        

    
    if returnAll:
        # Change to return average, std... for mean and SD of normal PDF.
        return logaverage, logstd, allMeasurements
    else:
        return logaverage, logstd

def area(mask, estSize, returnAll=False, showSteps=False):
    """
    Calculates the area of pores or seperated ligaments.
    Only meaningful if the phase is not fully connected.

    Parameters
    ----------
    mask : 2D array of bool
        Mask showing the area of one phase. 
    estSize : float64
        The estimated diameter.  Used to set some parameters.
    showSteps : bool, optional
        If set to true, figures will be displayed that shows the results of
        intermediate steps when calculating the diameter.
    returnAll : bool, optional
        If set to true, a list of all the measurements will be returned.

    Returns
    -------
    average : float64
        Average ojbect area.
    SD : float64
        Standard deviation of areas.
    areas: 1D array of float64, optional
        A list of all measured object areas.
    """
    
    # Return 0 if the mask is fully connected.
    if isConnected(mask):
        if returnAll:
            return 0,0,0
        else:
            return 0,0
        
    im = mask.copy()
    
    # Remove small areas that are probably just poking through the other phase.
    small = 3.14 * (estSize/2)**2
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        im = remove_small_objects(im, small)
    
    # label each blob.  0 = background, 1 = blobs touching edge.
    lig_labels = label(im, background=0)
    #subtract 1 for the ones connected to edge.
    num_blobs = np.amax(lig_labels) - 1 
    
    areas = np.bincount(lig_labels.flatten()) # Count pixels in each ligament.
    areas[1:][::-1].sort() # Sort the areas (ignore background).
    average = np.average(areas[2:]) # Ignore background and edge blobs.
    std_dev = np.std(areas[2:])

    #If more area of ligament is connected to the background than not,
    #then there is basically 100% in plane connectivity.
    #Report 0 in this case.
    if areas[1] > np.sum(areas[2:]):
        average = 0
        std_dev = 0
                         

    # animate labels
    if showSteps:
        fig = plt.figure()
        ims = []
        ims.append([plt.imshow(im, cmap=plt.get_cmap('gray'), animated=True)])
        ims.append([plt.imshow(lig_labels, animated=True)])
        ani = animation.ArtistAnimation(fig, ims, 1000, True, 1000)
        plt.axis('off')
        plt.show()

    if returnAll:
        return average, std_dev, areas[2:]
    else:
        return average, std_dev

def connectedLength(mask, estSize, returnAll=False, showSteps=False, pdf=None):
    """
    Calculates the average sum connected ligament length size and standard
    deviation of the passed of the passed mask.  This is the average length
    of the backbone of the objects in the mask when summing the length of each
    fork and branch in the object.

    Parameters
    ----------
    mask : 2D array of bool
        Mask showing the area of one phase.
    estSize : float64
        The estimated diameter.  Used to set some parameters.
    returnAll : bool, optional
        If set to true, a list of all the measurements will be returned.
    showSteps : bool, optional
        If set to true, figures will be displayed that shows the results of
        intermediate steps when calculating the diameter.
    pdf : matplotlib.backends.backend_pdf.PdfPages, optional
        If provided, information will be output to the pdf file.

    Returns
    -------
    average : float64
        Average sum length.
    SD : float64
        Standard deviation of sum lengths.
    lengths: 1D array of float64, optional
        A list of all measured sum lengths.
    """

    # Adjustable parameters.
    # These have been highly tested and changing these may throw off results.
    smooth_div = 5 # Sets the structure element radius for edge smoothing.
    
    im = mask.copy()
    
    # Return 0 if the mask is fully connected.
    if isConnected(mask):
        if returnAll:
            return 0,0,0
        else:
            return 0,0

    # Do some extra smoothing to the mask.
    smooth_rad = max(estSize/smooth_div, 1)
    im = segment.smoothEdges(im, smooth_rad)

    skel = skeleton(mask, estSize)
    dist = distanceTransform(mask)

    # Label each section.
    labels = label(im, connectivity=2)
    labels[labels==1] = 0 #Remove labels touching the edge
    labels = skel * labels

    # Get lengths.
    lengths = np.bincount(labels[labels>0].flatten())

    # When the skel backbone does not touch the edge of the object, add the
    # distance to the edge.
    # Find number of neighboring pixels.
    neighbors = convolve2d(skel, np.ones((3,3)), mode='same')
    neighbors = neighbors * skel
    term_neighbors = neighbors < 3
    term_neighbors[neighbors == 0] = 0    
    # If the endpoint of a ligament does not touch edge,
    # add the distance to the edge.
    for l in np.unique(labels):
        if l > 1:
            this_label = labels == l
            sum_places = term_neighbors*this_label
            to_sum = sum_places * dist
            lengths[l] += np.sum(to_sum)

    lengths = lengths[lengths>0]
    average = np.average(lengths)
    SD = np.std(lengths)

    if showSteps:
        display.showSkel(labels, mask, dialate=False)

    if showSteps:
        display.showHist(lengths, gauss=True, log=True,
                         title='Sum connected lengths',
                         xlabel='Object lengths [pixels]')
        
    if pdf is not None:
        inout.pdfSaveSkel(pdf, labels, mask, dialate=True)
        inout.pdfSaveHist(pdf, lengths, gauss=True, log=True,
                       title='Sum connected lengths',
                       xlabel='Object lengths [pixels]')
        
    if returnAll:
        return average, SD, lengths
    else:
        return average, SD
